// fir filter code
// filters a .wav file to a second .wav file using a specified filter
// coefficient .fir file

// usage: fir input.wav output.wav filter.fir

/*
#include "f2.inc"
#include "f3.inc"
#include "f4.inc"
#include "f5.inc"
#include "f6.inc"
#include "f7.inc"
#include "f8.inc"
#include "f9.inc"
#include "f10.inc"
#include "f11.inc"
#include "fx.inc"

#define coeffs 	FXCoeffs
#define taps    FXTaps
*/
#include "fx.inc"

#include <stdio.h>
#include <process.h>
#include <conio.h>
#include <alloc.h>
#include <dos.h>
#include <time.h>
#include <mem.h>
#include <string.h>

int fir(int params[], int numtaps, int input);

void fir2(int params[],
		int numtaps,
		int samples,
		int far * input,
		int far * output);

struct _waveheader {
	char 	riff[4];
	long 	block_size;
	char 	wave[4];
	char 	fmt[4];
	long 	chunk_size;
	int    	tag;
	int		channels;
	long	samples_per_sec;
	long	bytes_per_sec;
	int		block_align;
	int		bits_per_sample;
	char	data[4];
	long	data_size;
	};

#define BUFFSIZE 10000
#define MAXTAPS 101

_waveheader wh;              			// header for wave files
FILE * fi, * fo, * ff;					// file handles
//static int inbuf[BUFFSIZE];		// space for wave data
//static int outbuf[BUFFSIZE];   	// space for adpcm data
int far * inbuf;
int far * outbuf;
int far * inst;
int far * outst;
time_t first, second;

int main (int argc, char * argv[])
{
long count;						// number of bytes in block
long q;
int r;
int s;
int inumtaps;
static int itaps[MAXTAPS];


	if (argc < 4)
	{
		printf("Please specify 'from', 'to', and 'filter' files...\n");
		exit(1);
	}
	clrscr();
	printf("Opening:\n\t%s\n\t%s\n\t%s\nfor filtering\n",argv[1],argv[2],argv[3]);
	//getch();
	printf("Memory available: %ld\n",farcoreleft());

	fi = fopen(argv[1],"rb");	// read binary input data
	fread(&wh,sizeof(wh),1,fi);

	fo = fopen(argv[2],"wb");			// write binary as output
	fwrite (&wh,sizeof(wh),1,fo);		// write the header

	ff = fopen(argv[3],"rb");
	fread(&inumtaps,sizeof(int),1,ff);

	// a modicum of error checking...
	if ((fi == NULL)||(fo == NULL)||(ff == NULL))
	{
		printf("One of the specified files can't be opened!\n");
		fclose(fi);
		fclose(fo);
		fclose(ff);
		exit(1);
	}
	if ((strncmp(wh.riff,"RIFF",4) != 0) ||		// not a riff file
		(strncmp(wh.data,"data",4) != 0) ||     // no data
		(wh.bits_per_sample != 16) ||           // must be 16 bits
		(wh.channels != 1))                     // must be mono
	{
		printf("%s is in some way deficient - I can't use it.\n"
				"It should be a Windows wave file, mono, 16 bits per sample",
				argv[1]);
		fclose(fi);
		fclose(fo);
		fclose(ff);
		exit(1);
	}
	if (inumtaps==0)
	{
		printf("%s has no taps!\n",argv[3]);
		fclose(fi);
		fclose(fo);
		fclose(ff);
		exit(1);
	}
	if (inumtaps>MAXTAPS)
	{
		printf("s has too many taps!\n",argv[3]);
		fclose(fi);
		fclose(fo);
		fclose(ff);
		exit(1);
	}

	fread(itaps,sizeof(int),inumtaps,ff);

	q = 0;									// total bytes processed
	inst = (int far *)farmalloc(BUFFSIZE*2);
	outst = (int far *)farmalloc(BUFFSIZE*2);

	if ((inst == NULL) || (outst == NULL))
	{
		printf("Can't allocate sufficient memory.\n");
		fclose(fi);
		fclose(fo);
		fclose(ff);
		exit(1);
	}

	printf("Data size...%ld\nInput buffer %x:%x\nOutput buffer %x:%x\n",
			wh.data_size,FP_SEG(inst),FP_OFF(inst),FP_SEG(outst),FP_OFF(outst));
	//getch();
	first = time(NULL);

	while (q<wh.data_size)
	{
		if ((q+BUFFSIZE)<wh.data_size)
			count=BUFFSIZE;
		else
			count=wh.data_size-q;

		inbuf = inst;
		outbuf = outst;

		fread(inbuf,1,count,fi);
		fir2(itaps,inumtaps,count/2,inbuf,outbuf);
		//fir2(FXCoeffs,FXTaps,count/2,inbuf,outbuf);
		fwrite (outst,1,count,fo); // and stuff it to disk

		q += count;
		printf("+");
	}
	farfree(inst);
	farfree(outst);
	fclose(fi);
	fclose(fo);
	fclose(ff);
	second = time(NULL);

	printf("\n\nTime taken: %d seconds\n",(int)difftime(second,first));

	printf("Memory available: %ld\n",farcoreleft());
	//getch();
}

// perform an fir filter
// the input parameters are:
// a pointer to the delay block buffer
// a pointer to the filter parameters
// the number of taps
// the input sample
// the return value is the filter output

int fir(int params[], int numtaps, int input)
{
int q;
long total = 0;
int th,tl;
static int buffer[MAXTAPS];

	//printf("Params %p\nbuffer %p\n",params,buffer);
	//printf ("ds: %x\n",_DS);
	//getch();
	/*
	for (q=numtaps-1; q>0; q--)
	{
		// add the product to the total
		total += (long)buffer[q] * (long)params[q];
		// move the value up one...
		buffer[q] = buffer[q-1];
	}
	// then sort out the last entry
	total += (long)buffer[0]*(long)params[0];
	// insert the new value
	*/
	/*
	for (q=0; q<numtaps; q++)
		total += (long)buffer[q] * (long)params[q];
	_fmemcpy(&buffer[1],&buffer[0],numtaps-1);

	buffer[0] = input;
	// return the product
	return (total>>16);
	*/
	th = 0;
	tl = 0;
	_CX = numtaps;
	_SI = FP_OFF(params);
	_DI = FP_OFF(buffer);

	asm mov bx,cx
	asm sub bx,1
	asm shl bx,1				// preset the counter
loop1:
	asm mov ax,[si+bx]       	// do the multiply/add
	asm imul word ptr [di+bx]
	asm add word ptr tl,ax
	asm adc word ptr th,dx
	asm mov ax,[di+bx-2]        // shift the data
	asm mov[di+bx],ax
	asm sub bx,2
	asm loop loop1

	//_fmemcpy(&buffer[1],&buffer[0],numtaps-1);

	buffer[0] = input;
	// return the product
	return (th);


}
// this version moves the input and output parameters and the loop to the
// subroutine...it cuts down *lots* of calls and stack use
void fir2(int params[],
		int numtaps,
		int samples,
		int far * input,
		int far * output)
{
int th,tl,r,q;
static int buffer[63];
long total;

/*
//	this is the 'c' version...it's sloooow

	for (r=0; r<samples; r++)
	{
		total = 0;
		for (q=numtaps-1; q>=0; q--)
		{
			total += (long)buffer[q] * (long)params[q];
			buffer[q] = buffer[q-1];
		}

		buffer[0] = *input++;
		*output++ = (total>>16);
	}
*/

//	a quicker assembly version

	for (r=0; r<samples; r++)
	{
		th = 0;
		tl = 0;
		_CX = numtaps;
		_SI = FP_OFF(params);
		_DI = FP_OFF(buffer);

		asm mov bx,cx
		asm sub bx,1
		asm shl bx,1				// preset the counter
loop1:
		asm mov ax,[si+bx]       	// do the multiply/add
		asm imul word ptr [di+bx]
		asm add word ptr tl,ax
		asm adc word ptr th,dx
		asm mov ax,[di+bx-2]        // shift the data
		asm mov[di+bx],ax
		asm sub bx,2
		asm loop loop1

		buffer[0] = *input++;
		*output++ = th;
	}

}